@@ -70,7 +70,7 @@ method/sub declaration for the symbol (module/method/sub) at point. If
 no sub declaration is available (like for generated getters/setters),
 any appropriate POD is used instead.
 
-B<Go to Use> -- C<C-o g u> -- Go to the 'use Module' section of the current buffer.
+B<Go To Use> -- C<C-o g u> -- Go to the 'use Module' section of the current buffer.
 
 B<Go To 'new'> -- C<C-o g n> -- Go to the 'new' method of the current
 class.
@@ -101,11 +101,17 @@ of the method name, or word at point.
 B<Find method calls> -- C<C-o f c> -- Search for method calls of the
 method name, or word at point.
 
+B<Go To Find Buffer> -- C<C-o g f> to go to the B<*grep*> buffer.
+
 B<Run file> -- C<C-o C-r> -- Run the current file using the
 Compilation mode and the settings appropriate for the source type
 (Test, Module, etc.). Highlight errors and jump to source with C-c
 C-c.
 
+B<Edit - Add Use Statement> -- C<C-o e a u> -- Add a 'use Module'
+statement to the 'use Module' section at the top. Default Module name
+is module at point.
+
 B<Edit - Move Use Statement> -- C<C-o e m u> -- Move the 'use Module'
 statement at point to the 'use Module' section at the top.
 
@@ -173,7 +179,7 @@ Cache all modules in the project that --dir belongs to.
 Cache all the modules in @INC.
 
 This is a useful thing to do after installation (and after each
-upgrade), but it will take a wile so put it in the background and let
+upgrade), but it will take a while so put it in the background and let
 it churn away at those modules.
 
 =over 4
@@ -451,7 +457,10 @@ globally with C-x C-SPC, or locally with C-u C-SPC.
 =head3 Go to Module
 
 C<C-o g m> -- Go to Module at point. Useful if "Smart go to" can't
-identify what's at point.
+identify exactly what's at point.
+
+Default is the selected text, or the
+Module at point.
 
 
 =head3 Go to Base Class
@@ -579,6 +588,17 @@ you'll get to edit the command line with a sensible default chosen from:
 
 =back
 
+When editing the ack command you can use the following keys to set options
+
+ |---------+--------+--------+------------------------------------------|
+ | "C-o w" | toggle | -w     | Whole word                               |
+ | "C-o q" | toggle | -Q     | Quote metacharacters, pattern is literal |
+ | "C-o i" | toggle | -i     | Ignore case                              |
+ | "C-o a" | use    | --all  |                                          |
+ | "C-o p" | use    | --perl |                                          |
+ | "C-o s" | use    | --sql  |                                          |
+ |---------+--------+--------+------------------------------------------|
+
 For details, refer to the L<ack> documentation (the program was
 installed as a dependency of PerlySense).
 
@@ -620,6 +640,11 @@ word at point.
 I.e. look for lines with C<-E<gt>NAME>.
 
 
+=head3 Go to Find-buffer
+
+Invoke C<C-o g f> to go to the B<*grep*> buffer.
+
+
 
 =head2 Class Overview
 
@@ -758,10 +783,16 @@ C<C-o C-r> -- Run the file of the current buffer using the Compilation
 mode.
 
 Files are run according to the source type, which is determined by the
-file name (see the config file).  The default for .t files is to run
+file name (see the config file). The default for .t files is to run
 "prove -v", for .pm files "perl -c", etc. This can be configured per
 Project (see below).
 
+Files can also be run using an Alternate Command using C<C-u C-o C-r>
+if you have specified one in the config file. This might be useful if
+you want to re-generate or restart something before running the file,
+but only sometimes. Or, maybe you want to run some tests without the
+-v flag or something.
+
 The file is run from the Project root directory or from the file
 directory depending on the file type, and the @INC is set
 appropriately. You can also specify additional @INC directories in the
@@ -773,8 +804,10 @@ not just Perl source files.
 As a taste of what's possible, imagine that you have a test framework
 with .yml acceptance test data files and a corresponding yml-runner.pl
 script. You can set up the config so you can type C<C-o C-r> while
-editing the .yaml file to run that test. Refer to the
-L<Devel::PerlySense::Cookbook> for details.
+editing the .yaml file to run that test. And if you need to regenerate
+some fixtures or something before running the yml test, you can
+configure the Alternate Command to do that (run with C<C-u C-o
+C-r>). Refer to the L<Devel::PerlySense::Cookbook> for details.
 
 If any warnings, errors or test failures are encountered, they are
 highlighted in the B<*compilation*> buffer. Press RET on a highlighted
@@ -795,6 +828,9 @@ source fixing errors and the .t file isn't visible.
 C<C-o r r> -- If not even the B<*compilation*> buffer is visible,
 issue Re-Run File from anywhere to bring it up and re-run.
 
+Note: this will re-run whatever is displayed in the B<*compilation*>
+buffer.
+
 
 =head3 Go to Run-buffer
 
@@ -891,8 +927,12 @@ Note that if you have spaces in your file names, this might not work
 (it's a perldb thing).
 
 The debugger is started according to the file source type, which is
-determined by the file name (see the config file). This can be
-configured similar to how files are run (see above).
+determined by the file name (see the config file).
+
+You can also use C<C-u C-o r d> to Debug with an Alternate Command,
+just like with Run File.
+
+This can all be configured similar to how files are run (see above).
 
 Most files are run from the Project root directory by default.
 
@@ -1130,15 +1170,32 @@ I<that particular subroutine>.
 =head2 Editing Code
 
 
+=head3 Edit Add 'use Module' Statement
+
+C<C-o e a u> -- Set mark and add a 'use My::Module;' statement to the
+end of the 'use Module' section at the top of the file.
+
+The default module is the selected text, or the module at point (point
+may be on a method call of the module).
+
+This is typically useful when you realize you're using a module
+already, but without a use-statement. But you don't want to leave
+where you are just to fiddle with adding it.
+
+So hit C<C-o e a u> to add it, see that it got added at a good place
+and hit C-u C-SPC to return to where you were, and continue doing what
+you where doing.
+
+
+
 =head3 Edit Move 'use Module' Statement
 
 C<C-o e m u> -- If point is on a line with a single 'use Module'
 statement, set mark and move that statement to the end of the 'use
 Module' section at the top of the file.
 
-This is typically useful when you realize you need a module,
-e.g. Data::Dumper, in the middle of the file, but you don't want to
-leave where you are just to fiddle with adding it.
+This is typically useful for when you encounter a stray 'use Module'
+in the middle of the file.
 
 So type the 'use Module' statement, hit C<C-o e m u> to move it, see
 that it got moved to a good place and hit C-u C-SPC to return to where
@@ -1535,9 +1592,15 @@ L<http://vimdoc.sourceforge.net/htmldoc/eval.html#Dictionaries> - Vim native dat
 
 =head1 AUTHOR
 
-Johan Lindstrm, C<< <johanl[T]DarSerMan.com> >>
+Johan Lindstrm, C<< <johanl buzzwordninja.com> >>
+
+=head1 CONTRIBUTIONS, BUGS, AND CAVEATS
+
+=head2 CONTRIBUTIONS
+
+If you want to hack on PerlySense, fork the project at GitHub:
+L<https://github.com/jplindstrom/p5-Devel-PerlySense>
 
-=head1 BUGS AND CAVEATS
 
 =head2 BUG REPORTS
 
@@ -1596,7 +1659,9 @@ use strict;
 use warnings;
 
 package Devel::PerlySense;
-our $VERSION = '0.0187';
+BEGIN {
+  $Devel::PerlySense::VERSION = '0.0205'; # TRIAL
+}
 
 
 
@@ -1611,6 +1676,8 @@ use Pod::Text;
 use IO::String;
 use Cache::Cache;
 use Storable qw/freeze thaw/;
+use List::Util qw/ first /;
+use List::MoreUtils qw/ uniq /;
 
 use Devel::TimeThis;
 
@@ -1656,7 +1723,7 @@ Devel::PerlySense::Project object.
 Default: A Devel::PerlySense::Project::Unknown object.
 
 =cut
-field "oProject" => Devel::PerlySense::Project::Unknown->new();
+field "oProject" => undef;
 
 
 
@@ -1694,6 +1761,20 @@ sub rhConfig {
 
 
 
+=head2 VERSION
+
+The $VERSION of this module.
+
+=cut
+sub VERSION {
+    # This variable is created by Dist::Zilla during release
+    return $Devel::PerlySense::VERSION || "0.0001DEV";
+}
+
+
+
+
+
 =head2 oBookmarkConfig
 
 Devel::PerlySense::BookmarkConfig object.
@@ -1726,6 +1807,7 @@ Create new PerlySense object.
 sub new() {
     my $self = bless {}, shift;
     $self->oBookmarkConfig(Devel::PerlySense::BookmarkConfig->new( oPerlySense => $self ));
+    $self->oProject(Devel::PerlySense::Project::Unknown->new( oPerlySense => $self ));
     return($self);
 }
 
@@ -1780,6 +1862,14 @@ Die on errors (like if the file wasn't found).
 sub oDocumentParseFile {
 	my ($file) = @_;
 
+    # Stop recursive lookups
+    if( exists $self->rhFileDocumentCache->{$file}) {
+        if(! defined $self->rhFileDocumentCache->{$file}) {
+            die("Tried to parse ($file) recursively\n");
+        }
+    }
+    $self->rhFileDocumentCache->{$file} = undef;
+
     my $oDocument = $self->rhFileDocumentCache->{$file} ||= do {
         my $oDocumentNew = Devel::PerlySense::Document->new(oPerlySense => $self);
         $oDocumentNew->parse(file => $file);
@@ -2120,7 +2210,7 @@ sub raFileProjectOther {
 
 
 
-=head2 rhRunFile(file => $fileSource)
+=head2 rhRunFile(file => $fileSource, [ keyConfigCommand => "command" ])
 
 Figure out what type of source file $fileSource is, and how it should
 be run.
@@ -2144,14 +2234,14 @@ sub rhRunFile {
     $self->setFindProject(file => $file)
             or die("Could not identify any PerlySense Project\n");
 
-    return $self->oProject->rhRunFile(file => $file);
+    return $self->oProject->rhRunFile(@_);
 }
 
 
 
 
 
-=head2 rhDebugFile(file => $fileSource)
+=head2 rhDebugFile(file => $fileSource, [ keyConfigCommand => "command" ])
 
 Figure out what type of source file $fileSource is, and how it should
 be debugged.
@@ -2175,7 +2265,7 @@ sub rhDebugFile {
     $self->setFindProject(file => $file)
             or die("Could not identify any PerlySense Project\n");
 
-    return $self->oProject->rhDebugFile(file => $file);
+    return $self->oProject->rhDebugFile(@_);
 }
 
 
@@ -2233,6 +2323,7 @@ sub createProject {
 
     my $oConfig = Devel::PerlySense::Config::Project->new();
     $oConfig->createFileConfigDefault(dirRoot => $dir);
+    $oConfig->createFileCriticDefault(dirRoot => $dir);
 
     ###TODO: assign the config to $self->oConfigProject
 
@@ -2330,12 +2421,16 @@ errors.
 =cut
 sub fileFindModule {
     my ($nameModule, $dirOrigin) = Devel::PerlySense::Util::aNamedArg(["nameModule", "dirOrigin"], @_);
+
+    # TODO: Move this into fileFindLookingInInc and pass in the dir
+    $self->setFindProject(dir => $dirOrigin);
+
 #my $tt = Devel::TimeThis->new("fileFindModule");
     my $fileModuleBase = $self->fileFromModule($nameModule);
     $dirOrigin = dir($dirOrigin)->absolute;
 
     return(
-        $self->fileFindLookingAround($fileModuleBase, $dirOrigin) ||
+        $self->fileFindLookingAround($fileModuleBase, $dirOrigin, $nameModule) ||
         $self->fileFindLookingInInc($fileModuleBase) ||
         undef
     );
@@ -2396,22 +2491,68 @@ sub isFileInProject {
 
 =head1 IMPLEMENTATION METHODS
 
-=head2 fileFindLookingAround($fileModuleBase, $dirOrigin)
+=head2 fileFindLookingAround($fileModuleBase, $dirOrigin, $nameModule?)
 
-Find the file containing the $fileModuleBase given the $dirOrigin.
+Find the file containing the $fileModuleBase given the $dirOrigin. If
+$nameModule is specified, the file must either be in the inc_dir, or
+contain a package declaration for $nameModule.
 
 Return the file name relative to $dirOrigin, or undef if none could be
 found. Die on errors.
 
 =cut
 sub fileFindLookingAround {
-	my ($fileModuleBase, $dirOrigin) = @_;
+	my ($fileModuleBase, $dirOrigin, $nameModule) = @_;
+
+    my @aDirIncProject = map { dir($_)->absolute . "" }
+        $self->oProject->aDirIncProject(
+            dirRelativeTo => $self->oProject->dirProject,
+        );
 
     my $dir = dir($dirOrigin);
     while(1) {
         for my $dirCur (map { dir($dir, $_) } qw/. bin lib/) {
             if(my $fileFound = $self->fileFoundInDir($dirCur, $fileModuleBase)) {
-                return(file($fileFound)->absolute . "");
+                # is it in a local inc_dir?
+                if( first { $_ eq $dir } @aDirIncProject) {
+                    return(file($fileFound)->absolute . "");
+                }
+
+                # Are we expecting a module name? If not, it's a match.
+                $nameModule or return(file($fileFound)->absolute . "");
+
+
+                # Check for the dir above the file, is there a package
+                # name like that in the file? If so, this one isn't
+                # it.
+                # If I do this, the next one might not even be needed
+
+
+                # Does the file contain a Package declaration for the
+                # module name? This is a manual and cheap workaround
+                # to avoid recursive and slow parse
+                my $textFile = file($fileFound)->slurp();
+                if($textFile =~ m|
+                    package          # package declaration
+                    \s+
+                    [^;]*?           # up until until the next
+                                     # statement separator (fragile,
+                                     # could well be in comments or a
+                                     # block)
+                    (?<!::)          # Not preceeded by a module
+                                     # separator, i.e. it's not a
+                                     # module shadowing the shorter
+                                     # name
+                    $nameModule
+                    \b
+                    (?!::)           # Not followed by a module
+                                     # separator, i.e. it's not a
+                                     # longer, other module
+                |xsm) {
+                    ###TODO: possibly check using parse here, now that
+                    ###we know the package name is in there.
+                    return(file($fileFound)->absolute . "");
+                }
             }
         }
 
@@ -2464,16 +2605,19 @@ sub dirFindLookingAround {
 
 =head2 fileFindLookingInInc($fileModuleBase)
 
-Find the file containing the $nameModule in @INC.
+Find the file containing the $nameModule in config:project/extra_inc,
+and @INC.
 
 Return the absolute file name, or undef if none could be found. Die on
 errors.
 
 =cut
+
 sub fileFindLookingInInc {
 	my ($fileModuleBase) = @_;
 
-    for my $dirCur (@INC) {
+    my @aDirInc = uniq( $self->oProject->aDirIncAbsolute(), @INC );
+    for my $dirCur (@aDirInc) {
         if(my $fileFound = $self->fileFoundInDir($dirCur, $fileModuleBase)) {
             return($fileFound);
         }
@@ -2756,10 +2900,6 @@ sub cacheKeyTotal {
     return($keyTotal);
 }
 
-
-
-
-
 1;
 
 
